from collections import deque
from babel.messages.jslexer import tokenize, unquote_string


# This helper is copied to all necessary files because once again python imports are being a pain
def reopen_normal_read(file_obj, encoding):
    """Re-open a file obj in plain read mode"""
    return open(file_obj.name, "r", encoding=encoding)


JSON_GETTEXT_KEYWORD = 'type'
JSON_GETTEXT_VALUE = 'gettext_string'
JSON_GETTEXT_KEY_CONTENT = 'content'
JSON_GETTEXT_KEY_ALT_CONTENT = 'alt_content'
JSON_GETTEXT_KEY_FUNCNAME = 'funcname'


class JsonExtractor(object):
    def __init__(self, data):
        self.state = 'start'
        self.data = data

        self.token_to_add = None
        self.is_value = False
        self.gettext_mode = False
        self.current_key = None
        self.in_array = False
        self.nested_in_array = []
        self.results = []
        self.token_params = {}

    # TODO: fix the duplicate name between this and the other add_result
    def add_result(self, token):
        value = unquote_string(token.value)
        if value not in self.results:
            self.results[value] = deque()
        self.results[value].append(token.lineno)

    def start_object(self):
        self.gettext_mode = False
        self.state = 'key'

        # Things will be incorrect if an object is contained in an array, so
        # we use a stack of states to return to like this in order to support
        # that kind of JSON structures
        self.nested_in_array.append(self.in_array)
        self.in_array = False

    def with_separator(self, token):
        self.state = 'value'

    def start_array(self):
        self.in_array = True

    def end_array(self):
        self.in_array = False
        self.end_pair()

    def end_pair(self, add_gettext_object=False):
        if self.token_to_add:
            if not self.gettext_mode or (self.gettext_mode and add_gettext_object):
                self.add_result(self.token_to_add)

        if not self.in_array:
            self.current_key = None
            self.state = 'key'

    def end_object(self):
        self.end_pair(add_gettext_object=True)
        self.gettext_mode = False
        self.state = 'end'

        self.in_array = self.nested_in_array.pop()

    def add_result(self, token):
        value = unquote_string(token.value)
        result = dict(
            line_number=token.lineno,
            content=value
        )
        for key, value in self.token_params.items():
            if key == 'alt_token':
                result['alt_content'] = unquote_string(value.value)
                result['alt_line_number'] = value.lineno
            else:
                result[key] = unquote_string(value)

        self.results.append(result)
        self.token_to_add = None
        self.token_params = {}

    def get_lines_data(self):
        """
        Returns string:line_numbers list
        Since all strings are unique it is OK to get line numbers this way.
        Since same string can occur several times inside single .json file the values should be popped(FIFO) from the list
        :rtype: list
        """

        for token in tokenize(self.data):
            if token.type == 'operator':
                if token.value == '{':
                    self.start_object()
                elif token.value == '[':
                    self.start_array()
                elif token.value == ':':
                    self.with_separator(token)
                elif token.value == '}':
                    self.end_object()
                elif token.value == ']':
                    self.end_array()
                elif token.value == ',':
                    self.end_pair()


            elif token.type == 'string':
                if self.state == 'key':
                    self.current_key = unquote_string(token.value)
                    if self.current_key == JSON_GETTEXT_KEYWORD:
                        self.gettext_mode = True
                else:
                    # TODO: auto-detecting items to extract through the keywords passed to extract_json would be very nice
                    if self.current_key.lower() in (
                    "groupname", "displayname", "name", "message", "messages"):
                        self.token_to_add = token

        return self.results


def extract_json(file_obj, keywords, comment_tags, options):
    """
    Supports: gettext, ngettext. See package README or github ( https://github.com/tigrawap/pybabel-json ) for more usage info.
    """
    with reopen_normal_read(file_obj, options.get('encoding', 'utf-8')) as f:
        data = f.read()
    json_extractor = JsonExtractor(data)
    strings_data = json_extractor.get_lines_data()

    for item in strings_data:
        messages = [item['content']]
        if item.get('funcname') == 'ngettext':
            messages.append(item['alt_content'])
        yield item['line_number'], item.get('funcname', 'gettext'), tuple(messages), []



# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name='Babel-Thrive',
    version='1.11',
    description='Plugin for Babel to support Godot scene files (.tscn) extended for Thrive',
    author='Remi Rampin (original), Nicolas Viseur and Henri Hyyryläinen (fork)',
    author_email='westtunger@gmail.com',
    license='BSD',
    url='https://github.com/Revolutionary-Games/pybabel-godot-thrive',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    py_modules=['babel_godot', 'json_extractor', 'csharp_extractor'],

    entry_points="""
    [babel.extractors]
    godot_scene = babel_godot:extract_godot_scene
    godot_resource = babel_godot:extract_godot_resource
    json = json_extractor:extract_json
    csharp = csharp_extractor:extract_csharp
    """
)



import re

__version__ = '1.0'

_godot_node = re.compile(r'^\[node name="([^"]+)" (?:type="([^"]+)")?')
_godot_property_str = re.compile(r'^([A-Za-z0-9_]+)\s*=\s*(".+)$')
_string_number = re.compile(r'^[\+\-\*\/\%\.\s]*\d*[\+\-\*\/\%\.\s]*\d*[\+\-\*\/\%\.\s]*\d*$')
_godot_option_button = re.compile(r'^(items)\s*=\s*\[\s*(.+)\s*\]\s*$')

def _godot_unquote(string):
    result = []
    escaped = False
    for i, c in enumerate(string):
        if escaped:
            if c == '\\':
                result.append('\\')
            elif c == 'n':
                result.append('\n')
            elif c == 't':
                result.append('\t')
            else:
                result.append(c)
        else:
            if c == '\\':
                escaped = True
            elif c == '"':
                return ''.join(result), string[i + 1:]
            else:
                result.append(c)
    return ''.join(result), None


def extract_godot_scene(fileobj, keywords, comment_tags, options):
    """Extract messages from Godot scene files (.tscn).

    :param fileobj: the seekable, BufferedReader object the messages should be
                    extracted from
    :param keywords: a list of property names that should be localized, in the
                     format '<NodeType>/<name>' or '<name>' (example:
                     'Label/text')
    :param comment_tags: a list of translator tags to search for and include
                         in the results (ignored)
    :param options: a dictionary of additional options (optional)
    :rtype: iterator
    """
    encoding = options.get('encoding', 'utf-8')

    current_node_type = None

    properties_to_translate = {}
    for keyword in keywords:
        if '/' in keyword:
            properties_to_translate[tuple(keyword.split('/', 1))] = keyword
        else:
            properties_to_translate[(None, keyword)] = keyword

    def check_translate_property(property):
        keyword = properties_to_translate.get((current_node_type, property))
        if keyword is None:
            keyword = properties_to_translate.get((None, property))
        return keyword

    lines = []
    current_string = keyword = None

    for lineno, line in enumerate(fileobj, start=1):
        lines.append(line)

    for lineno, line in enumerate(lines):
        line = line.decode(encoding)

        if current_string:
            value, remainder = _godot_unquote(line)
            current_string.append(value)
            if remainder is None:  # Still un-terminated
                pass
            elif remainder.strip():
                raise ValueError("Trailing data after string")
            else:
                yield (lineno + 1, keyword, ['\n'.join(current_string)], [])
                current_string = None
            continue

        match = _godot_node.match(line)

        if match:
            # Store which kind of node we're in
            current_node_type = match.group(2)
            # instanced packed scenes don't have the type field,
            # change current_node_type to empty string
            current_node_type = current_node_type if current_node_type is not None else ""
        elif line.startswith('['):
            # We're no longer in a node
            current_node_type = None
        elif current_node_type is not None:
            # Currently in a node, check properties
            match = _godot_property_str.match(line)
            if match:
                if check_for_placeholder(lines[lineno:], encoding):
                    continue

                property = match.group(1)
                value = match.group(2)
                keyword = check_translate_property(property)
                if keyword:
                    value, remainder = _godot_unquote(value[1:])
                    if remainder is None:  # Un-terminated string
                        current_string = [value]
                    elif not remainder.strip():
                        if _string_number.match(value):
                            continue
                        yield (lineno + 1, keyword, [value], [])
                    else:
                        raise ValueError("Trailing data after string")
            else:
                # Options button items line handling
                match = _godot_option_button.match(line)
                if match:
                    if check_for_placeholder(lines[lineno:], encoding):
                        continue

                    raw_values = list(map(str.strip, match.group(2).split(",")))

                    # The format seems to that each option entry is 5 items
                    if len(raw_values) % 5 != 0:
                        print(f"not divisible (size: {len(raw_values)}: {raw_values}")
                        continue

                    for i in range(0, len(raw_values), 5):
                        option_entry = raw_values[i:i+5]

                        if option_entry[0][0] == '"' or option_entry[0][0] == '\'':
                            value = option_entry[0][1:-1]
                        else:
                            value = option_entry[0]

                        if _string_number.match(value):
                            continue

                        yield (lineno + 1, keyword, [value], [])


def check_for_placeholder(lines, encoding):
    for line in lines:
        line = line.decode(encoding)
        if 'PLACEHOLDER' in line:
            return True
        elif line.startswith('['):
            return False


def extract_godot_resource(fileobj, keywords, comment_tags, options):
    """Extract messages from Godot resource files (.res, .tres).

    :param fileobj: the seekable, file-like object the messages should be
                    extracted from
    :param keywords: a list of property names that should be localized, in the
                     format 'Resource/<name>' or '<name>' (example:
                     'Resource/text')
    :param comment_tags: a list of translator tags to search for and include
                         in the results (ignored)
    :param options: a dictionary of additional options (optional)
    :rtype: iterator
    """
    encoding = options.get('encoding', 'utf-8')

    properties_to_translate = {}
    for keyword in keywords:
        if keyword.startswith('Resource/'):
            properties_to_translate[keyword[9:]] = keyword

    def check_translate_property(property):
        return properties_to_translate.get(property)

    current_string = keyword = None

    for lineno, line in enumerate(fileobj, start=1):
        line = line.decode(encoding)

        if current_string:
            value, remainder = _godot_unquote(line)
            current_string.append(value)
            if remainder is None:  # Still un-terminated
                pass
            elif remainder.strip():
                raise ValueError("Trailing data after string")
            else:
                yield (lineno + 1, keyword, ['\n'.join(current_string)], [])
                current_string = None
            continue

        if line.startswith('['):
            continue

        match = _godot_property_str.match(line)
        if match:
            property = match.group(1)
            value = match.group(2)
            keyword = check_translate_property(property)
            if keyword:
                value, remainder = _godot_unquote(value[1:])
                if remainder is None:  # Un-terminated string
                    current_string = [value]
                elif not remainder.strip():
                    yield (lineno + 1, keyword, [value], [])
                else:
                    raise ValueError("Trailing data after string")
